import 'package:riverpod/riverpod.dart';
import 'package:endlich/endlich.dart';

import 'builder.dart';
import 'value_provider.dart';

mixin MachineStateProviderMixin<Value, Event>
    on ProviderBase<MachineState<Value, Event>> {
  ProviderBase<Service<Value, Event>> get service;

  @override
  ProviderBase<Object?> get providerToRefresh => service;

  @override
  void setupOverride(SetupOverride setup) {
    setup(origin: this, override: this);
    setup(origin: service, override: service);
  }

  Override overrideWithValue(Service<Value, Event> value) {
    return ProviderOverride((setup) {
      setup(
        origin: service,
        override: ValueProvider<Service<Value, Event>>(value),
      );
      setup(origin: this, override: this);
    });
  }
}

abstract class ServiceProviderRef<Value, Event> implements ProviderRefBase {}

class ServiceProviderElement<Value, Event>
    extends ProviderElementBase<Service<Value, Event>>
    implements ServiceProviderRef<Value, Event> {
  ServiceProviderElement(ProviderBase<Service<Value, Event>> provider)
      : super(provider);
}

class ServiceProvider<Value, Event>
    extends AlwaysAliveProviderBase<Service<Value, Event>> {
  ServiceProvider(this._create, {String? name})
      : super(name == null ? null : '$name.service');

  final Create<Service<Value, Event>, ServiceProviderRef<Value, Event>> _create;

  @override
  Service<Value, Event> create(ServiceProviderRef<Value, Event> ref) {
    final service = _create(ref);
    ref.onDispose(service.dispose);
    return service;
  }

  @override
  ServiceProviderElement<Value, Event> createElement() {
    return ServiceProviderElement<Value, Event>(this);
  }

  @override
  void setupOverride(SetupOverride setup) {
    throw UnsupportedError('Cannot override MachineStateProvider.service');
  }

  @override
  bool updateShouldNotify(
      Service<Value, Event> previousState, Service<Value, Event> newState) {
    return true;
  }
}

abstract class MachineStateProviderRef<Value, Event>
    implements ProviderRefBase {}

class MachineStateProviderElement<Value, Event>
    extends ProviderElementBase<MachineState<Value, Event>>
    implements MachineStateProviderRef<Value, Event> {
  MachineStateProviderElement(ProviderBase<MachineState<Value, Event>> provider)
      : super(provider);
}

class MachineStateProvider<Value, Event>
    extends AlwaysAliveProviderBase<MachineState<Value, Event>>
    with MachineStateProviderMixin<Value, Event> {
  MachineStateProvider(this._create, {String? name}) : super(name);

  static const family = MachineStateProviderFamilyBuilder();

  static const autoDispose = AutoDisposeMachineStateProviderBuilder();

  static const autoDisposeFamily =
      AutoDisposeMachineStateProviderFamilyBuilder();

  final Create<Service<Value, Event>, ServiceProviderRef<Value, Event>> _create;

  @override
  late final ServiceProvider<Value, Event> service =
      ServiceProvider<Value, Event>(_create, name: name);

  @override
  MachineState<Value, Event> create(MachineStateProviderRef<Value, Event> ref) {
    final ele = ref as MachineStateProviderElement<Value, Event>;
    final s = ref.watch(service);
    ref.onDispose(s.watch(
      (value) => ele.state = value,
      fireImmediately: true,
    ));
    return ele.state;
  }

  @override
  ProviderElementBase<MachineState<Value, Event>> createElement() {
    return MachineStateProviderElement<Value, Event>(this);
  }

  @override
  bool updateShouldNotify(MachineState<Value, Event> previousState,
      MachineState<Value, Event> newState) {
    return true;
  }

  Override overrideWithProvider(
    MachineStateProvider<Value, Event> provider,
  ) {
    return ProviderOverride((setup) {
      setup(origin: service, override: provider.service);
      setup(origin: this, override: this);
    });
  }
}

class MachineStateProviderFamily<Value, Event, Arg> extends Family<
    MachineState<Value, Event>, Arg, MachineStateProvider<Value, Event>> {
  MachineStateProviderFamily(this._create, {String? name}) : super(name);

  final FamilyCreate<Service<Value, Event>, ServiceProviderRef<Value, Event>,
      Arg> _create;

  @override
  MachineStateProvider<Value, Event> create(
    Arg argument,
  ) {
    final provider = MachineStateProvider<Value, Event>(
      (ref) => _create(ref, argument),
      name: name,
    );

    registerProvider(provider.service, argument);

    return provider;
  }

  /// Overrides the behavior of a family for a part of the application.
  ///
  /// {@macro riverpod.overideWith}
  Override overrideWithProvider(
    MachineStateProvider<Value, Event> Function(Arg argument) override,
  ) {
    return FamilyOverride<Arg>(
      this,
      (arg, setup) {
        final provider = call(arg);
        setup(origin: provider.service, override: override(arg).service);
        setup(origin: provider, override: provider);
      },
    );
  }

  @override
  void setupOverride(Arg argument, SetupOverride setup) {
    final provider = call(argument);
    setup(origin: provider, override: provider);
    setup(origin: provider.service, override: provider.service);
  }
}
