import 'package:endlich/endlich.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:riverpod/riverpod.dart';
import 'package:riverpod_endlich/src/provider.dart';
import 'package:test/test.dart';

import 'auto_dispose_test.mocks.dart';
import 'stuff.dart';

@GenerateMocks([StateListener])
void main() {
  test('can be refreshed', () async {
    var result = TestService();
    final container = createContainer();
    final provider =
        MachineStateProvider.autoDispose<TestValue, TestEvent>((ref) => result);

    expect(container.read(provider), isAValueOf(TestValue.foo()));
    expect(container.read(provider.service), result);

    result = TestService(initial: TestValue.bar());
    expect(container.refresh(provider), isAValueOf(TestValue.bar()));

    expect(container.read(provider), isAValueOf(TestValue.bar()));
    expect(container.read(provider.service), result);
  });

  group('scoping an override overrides all the associated subproviders', () {
    test('when passing the provider itself', () async {
      final service = TestService();
      final provider = MachineStateProvider.autoDispose<TestValue, TestEvent>(
          (ref) => service);
      final root = createContainer();
      final container = createContainer(parent: root, overrides: [provider]);

      expect(container.read(provider.service), service);
      expect(container.read(provider), isAValueOf(TestValue.foo()));
      expect(root.getAllProviderElements(), isEmpty);
      expect(
        container.getAllProviderElements(),
        unorderedEquals(<Object?>[
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider),
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider.service),
        ]),
      );
    });

    test('when using provider.overrideWithValue', () async {
      final service = TestService();
      final provider = MachineStateProvider.autoDispose<TestValue, TestEvent>(
          (ref) => service);
      final root = createContainer();
      final controllerOverride = TestService(initial: TestValue.bar());
      final container = createContainer(parent: root, overrides: [
        provider.overrideWithValue(controllerOverride),
      ]);

      expect(container.read(provider.service), controllerOverride);
      expect(container.read(provider), isAValueOf(TestValue.bar()));
      expect(root.getAllProviderElements(), isEmpty);
      expect(
        container.getAllProviderElements(),
        unorderedEquals(<Object?>[
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider),
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider.service),
        ]),
      );
    });

    test('when using provider.overrideWithProvider', () async {
      final service = TestService();
      final provider = MachineStateProvider.autoDispose<TestValue, TestEvent>(
          (ref) => service);
      final root = createContainer();
      final controllerOverride = TestService(initial: TestValue.bar());
      final container = createContainer(parent: root, overrides: [
        provider.overrideWithProvider(
          MachineStateProvider.autoDispose((ref) => controllerOverride),
        ),
      ]);

      expect(container.read(provider.service), controllerOverride);
      expect(container.read(provider), isAValueOf(TestValue.bar()));
      expect(root.getAllProviderElements(), isEmpty);
      expect(
        container.getAllProviderElements(),
        unorderedEquals(<Object?>[
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider),
          isA<ProviderElementBase>()
              .having((e) => e.origin, 'origin', provider.service),
        ]),
      );
    });
  });

  test('StateNotifierFamily override', () async {
    final notifier2 = TestService(initial: TestValue.bar());
    final provider =
        MachineStateProvider.autoDisposeFamily<TestValue, TestEvent, int>(
            (ref, a) => TestService());
    final container = createContainer(
      overrides: [
        provider.overrideWithProvider((a) {
          return MachineStateProvider.autoDispose<TestValue, TestEvent>(
            (ref) => notifier2,
          );
        }),
      ],
    );
    addTearDown(container.dispose);
    final ownerStateListener = MockStateListener();
    final ownerNotifierListener = ControllerListenerMock();

    // access in the child container
    // try to read provider.state before provider and see if it points to the override
    final stateSub = container.listen(
      provider(0),
      ownerStateListener,
      fireImmediately: true,
    );

    verify(ownerStateListener(argThat(isAValueOf(TestValue.bar())))).called(1);
    verifyNoMoreInteractions(ownerStateListener);

    final notifierSub = container.listen(
      provider(0).service,
      ownerNotifierListener,
      fireImmediately: true,
    );
    verifyOnly(ownerNotifierListener, ownerNotifierListener(notifier2));

    notifierSub.close();

    await container.pump();

    expect(notifier2.mounted, true);

    stateSub.close();

    expect(notifier2.mounted, true);

    await container.pump();

    await container.pump();

    expect(notifier2.mounted, false);
  });

  test('overriding the provider overrides provider.state too', () {
    final provider =
        MachineStateProvider.autoDispose<TestValue, TestEvent>((_) {
      return TestService();
    });
    final service = TestService(initial: TestValue.bar());
    final container = createContainer(
      overrides: [
        provider.overrideWithValue(TestService(initial: TestValue.baz())),
      ],
    );
    addTearDown(container.dispose);
    final stateListener = MockStateListener();
    final notifierListener = ControllerListenerMock();

    // does not crash
    container.updateOverrides([
      provider.overrideWithValue(service),
    ]);

    container.listen(provider.service, notifierListener, fireImmediately: true);
    verify(notifierListener(service)).called(1);
    verifyNoMoreInteractions(notifierListener);

    container.listen(provider, stateListener, fireImmediately: true);
    verify(stateListener(argThat(isAValueOf(TestValue.bar())))).called(1);
    verifyNoMoreInteractions(stateListener);

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.baz()));

    verify(stateListener(argThat(isAValueOf(TestValue.baz())))).called(1);
    verifyNoMoreInteractions(notifierListener);
    verifyNoMoreInteractions(stateListener);
  });

  test('can specify name', () {
    final provider = MachineStateProvider.autoDispose(
      (_) => TestService(),
      name: 'example',
    );

    expect(provider.service.name, 'example.service');
    expect(provider.name, 'example');

    final provider2 = MachineStateProvider.autoDispose((_) => TestService());

    expect(provider2.service.name, isNull);
    expect(provider2.name, isNull);
  });

  test('disposes the notifier when provider is unmounted', () {
    final service = TestService();
    final provider =
        MachineStateProvider.autoDispose<TestValue, TestEvent>((_) {
      return service;
    });
    final container = createContainer();
    addTearDown(container.dispose);

    container.listen(provider, (value) {});
    expect(service.mounted, isTrue);

    container.dispose();

    expect(service.mounted, isFalse);
  });

  test('provider subscribe the callback is never', () async {
    final service = TestService();
    final provider =
        MachineStateProvider.autoDispose<TestValue, TestEvent>((_) {
      return service;
    });
    final listener = ControllerListenerMock();
    final container = createContainer();
    addTearDown(container.dispose);

    container.listen(provider.service, listener, fireImmediately: true);

    verifyOnly(listener, listener(argThat(isA<TestService>())));

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.bar()));

    await container.pump();

    verifyNoMoreInteractions(listener);

    container.dispose();

    verifyNoMoreInteractions(listener);
  });

  test('provider subscribe callback never called', () async {
    final service = TestService();
    final provider =
        MachineStateProvider.autoDispose<TestValue, TestEvent>((_) {
      return service;
    });
    final listener = MockStateListener();
    final container = createContainer();
    addTearDown(container.dispose);

    container.listen(provider, listener, fireImmediately: true);

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.foo()))));

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.bar()));

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.bar()))))
        .called(1);

    container.dispose();

    verifyNoMoreInteractions(listener);
  });

  test('.notifier obtains the controller without listening to it', () async {
    final dep = StateProvider((ref) => 0);
    final service = TestService();
    final service2 = TestService();
    final provider =
        MachineStateProvider.autoDispose<TestValue, TestEvent>((ref) {
      return ref.watch(dep).state == 0 ? service : service2;
    });
    final container = createContainer();
    addTearDown(container.dispose);

    var callCount = 0;
    final sub = container.listen(provider.service, (_) => callCount++);

    expect(sub.read(), service);
    expect(callCount, 0);

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.bar()));

    expect(callCount, 0);

    container.read(dep).state++;

    expect(sub.read(), service2);

    await container.pump();

    expect(sub.read(), service2);
    expect(callCount, 1);
  });

  test('overrideWithProvider preserves the state accross update', () async {
    final provider =
        MachineStateProvider.autoDispose<TestValue, TestEvent>((_) {
      return TestService();
    });
    final service = TestService(initial: TestValue.bar());
    final service2 = TestService(initial: TestValue.baz());
    final container = createContainer(overrides: [
      provider.overrideWithProvider(
        MachineStateProvider.autoDispose<TestValue, TestEvent>((_) {
          return service;
        }),
      ),
    ]);
    addTearDown(container.dispose);
    final listener = MockStateListener();

    container.listen<MachineState<TestValue, TestEvent>>(provider, listener,
        fireImmediately: true);

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.bar()))));
    expect(container.read(provider.service), service);
    expect(service.hasListener, true);

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.foo()));

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.foo()))));

    container.updateOverrides([
      provider.overrideWithProvider(
        MachineStateProvider.autoDispose<TestValue, TestEvent>((_) {
          return service2;
        }),
      ),
    ]);

    await container.pump();

    expect(container.read(provider.service), service);
    expect(service2.hasListener, false);
    verifyNoMoreInteractions(listener);

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.bar()));

    await container.pump();

    expect(container.read(provider.service), service);
    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.bar()))));
    expect(service.mounted, true);

    container.dispose();

    expect(service.mounted, false);
  });
}
