import 'package:endlich/endlich.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:riverpod/riverpod.dart';
import 'package:riverpod_endlich/src/provider.dart';
import 'package:test/test.dart';

import 'riverpod_endlich_test.mocks.dart';
import 'stuff.dart';

@GenerateMocks([StateListener])
void main() {
  test('can be refreshed', () async {
    var result = TestService();
    final container = createContainer();
    final provider =
        MachineStateProvider<TestValue, TestEvent>((ref) => result);

    expect(container.read(provider), isAValueOf(TestValue.foo()));
    expect(container.read(provider.service), result);

    result = TestService(initial: TestValue.bar());
    expect(container.refresh(provider), isAValueOf(TestValue.bar()));

    expect(container.read(provider), isAValueOf(TestValue.bar()));
    expect(container.read(provider.service), result);
  });

  group('scope', () {
    group('scoping an override overrides all the associated subproviders', () {
      test('when passing the provider itself', () async {
        final service = TestService();
        final provider =
            MachineStateProvider<TestValue, TestEvent>((ref) => service);
        final root = createContainer();
        final container = createContainer(parent: root, overrides: [provider]);

        expect(container.read(provider.service), service);
        expect(container.read(provider), isAValueOf(TestValue.foo()));
        expect(root.getAllProviderElements(), isEmpty);
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.service),
          ]),
        );
      });

      test('when using provider.overrideWithValue', () async {
        final service = TestService();
        final provider =
            MachineStateProvider<TestValue, TestEvent>((ref) => service);
        final root = createContainer();
        final serviceOverride = TestService(initial: TestValue.bar());
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithValue(serviceOverride),
        ]);

        expect(container.read(provider.service), serviceOverride);
        expect(container.read(provider), isAValueOf(TestValue.bar()));
        expect(root.getAllProviderElements(), isEmpty);
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.service),
          ]),
        );
      });

      test('when using provider.overrideWithProvider', () async {
        final service = TestService();
        final provider =
            MachineStateProvider<TestValue, TestEvent>((ref) => service);
        final root = createContainer();
        final serviceOverride = TestService(initial: TestValue.bar());
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithProvider(
            MachineStateProvider<TestValue, TestEvent>(
                (ref) => serviceOverride),
          ),
        ]);

        expect(container.read(provider.service), serviceOverride);
        expect(container.read(provider), isAValueOf(TestValue.bar()));
        expect(root.getAllProviderElements(), isEmpty);
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.service),
          ]),
        );
      });
    });
  });

  test('MachineStateProviderFamily override', () {
    final provider =
        MachineStateProvider.family<TestValue, TestEvent, int>((ref, a) {
      return TestService();
    });
    final service2 = TestService(initial: TestValue.bar());
    final container = createContainer(
      overrides: [
        provider.overrideWithProvider((a) {
          return MachineStateProvider<TestValue, TestEvent>(
            (ref) => service2,
          );
        }),
      ],
    );
    addTearDown(container.dispose);

    // try to read provider.state before provider and see if it points to the override
    expect(container.read(provider(0)), isAValueOf(TestValue.bar()));
    expect(container.read(provider(0).service), service2);
  });

  test('overriding the provider overrides provider.state too', () {
    final service = TestService(initial: TestValue.bar());
    final provider =
        MachineStateProvider<TestValue, TestEvent>((_) => TestService());
    final container = createContainer(
      overrides: [
        provider.overrideWithValue(TestService(initial: TestValue.baz())),
      ],
    );
    addTearDown(container.dispose);

    // does not crash
    container.updateOverrides([
      provider.overrideWithValue(service),
    ]);

    expect(container.read(provider.service), service);
    expect(container.read(provider), isAValueOf(TestValue.bar()));

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.baz()));

    expect(container.read(provider), isAValueOf(TestValue.baz()));
  });

  test('can specify name', () {
    final provider = MachineStateProvider<TestValue, TestEvent>((_) {
      return TestService();
    }, name: 'example');

    expect(provider.service.name, 'example.service');
    expect(provider.name, 'example');

    final provider2 =
        MachineStateProvider<TestValue, TestEvent>((_) => TestService());

    expect(provider2.service.name, isNull);
    expect(provider2.name, isNull);
  });

  test('disposes the notifier when provider is unmounted', () {
    final service = TestService();
    final provider = MachineStateProvider<TestValue, TestEvent>((_) {
      return service;
    });
    final container = createContainer();
    addTearDown(container.dispose);

    expect(container.read(provider.service), service);
    expect(service.mounted, isTrue);

    container.dispose();

    expect(service.mounted, isFalse);
  });

  test('provider subscribe the callback is never', () async {
    final service = TestService();
    final provider = MachineStateProvider<TestValue, TestEvent>((_) {
      return service;
    });
    final listener = ControllerListenerMock();
    final container = createContainer();
    addTearDown(container.dispose);

    container.listen(provider.service, listener, fireImmediately: true);

    verifyOnly(listener, listener(argThat(isA<TestService>())));

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.bar()));

    await container.pump();

    verifyNoMoreInteractions(listener);

    container.dispose();

    verifyNoMoreInteractions(listener);
  });

  test('provider subscribe callback never called', () async {
    final service = TestService();
    final listener = MockStateListener();
    final provider = MachineStateProvider<TestValue, TestEvent>((_) {
      return service;
    });

    final container = createContainer();
    addTearDown(container.dispose);

    container.listen(provider, listener, fireImmediately: true);

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.foo()))));

    container
        .read(provider.service)
        .sendMachineEvent(MachineEvent.iToValue(value: TestValue.bar()));

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.bar()))));

    container.dispose();

    verifyNoMoreInteractions(listener);
  });

  test('.notifier obtains the controller without listening to it', () async {
    final dep = StateProvider((ref) => 0);
    final service = TestService();
    final service2 = TestService();
    final provider = MachineStateProvider<TestValue, TestEvent>((ref) {
      return ref.watch(dep).state == 0 ? service : service2;
    });
    final container = createContainer();
    addTearDown(container.dispose);

    var callCount = 0;
    final sub = container.listen(
      provider.service,
      (_) => callCount++,
    );

    expect(sub.read(), service);
    expect(callCount, 0);

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.bar()));

    await container.pump();

    expect(callCount, 0);

    container.read(dep).state++;

    expect(sub.read(), service2);

    await container.pump();

    expect(sub.read(), service2);
    expect(callCount, 1);
  });

  test(
      'overrideWithValue listens to the notifier, support notifier change, and does not dispose of the notifier',
      () async {
    final provider = MachineStateProvider<TestValue, TestEvent>((_) {
      return TestService();
    });
    final service = TestService(initial: TestValue.bar());
    final service2 = TestService();
    final container = createContainer(overrides: [
      provider.overrideWithValue(service),
    ]);
    addTearDown(container.dispose);
    final listener = MockStateListener();

    container.listen<MachineState<TestValue, TestEvent>>(provider, listener,
        fireImmediately: true);

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.bar()))));
    expect(container.read(provider.service), service);
    expect(service.hasListener, true);

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.baz()));

    await container.pump();

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.baz()))));

    container.updateOverrides([
      provider.overrideWithValue(service2),
    ]);

    await container.pump();

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.foo()))));

    expect(service.hasListener, false);
    expect(service.mounted, true);
    expect(container.read(provider.service), service2);

    service2.sendMachineEvent(MachineEvent.iToValue(value: TestValue.bar()));

    await container.pump();

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.bar()))));
  });

  test('overrideWithProvider preserves the state accross update', () async {
    final provider = MachineStateProvider<TestValue, TestEvent>((_) {
      return TestService();
    });
    final service = TestService(initial: TestValue.baz());
    final service2 = TestService(initial: TestValue.bar());
    final container = createContainer(overrides: [
      provider.overrideWithProvider(
        MachineStateProvider<TestValue, TestEvent>((_) {
          return service;
        }),
      ),
    ]);
    addTearDown(container.dispose);
    final listener = MockStateListener();

    container.listen<MachineState<TestValue, TestEvent>>(provider, listener,
        fireImmediately: true);

    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.baz()))));
    expect(container.read(provider.service), service);
    expect(service.hasListener, true);

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.foo()));

    await container.pump();
    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.foo()))));

    container.updateOverrides([
      provider.overrideWithProvider(
        MachineStateProvider<TestValue, TestEvent>((_) {
          return service2;
        }),
      ),
    ]);

    await container.pump();
    expect(container.read(provider.service), service);
    expect(service2.hasListener, false);
    verifyNoMoreInteractions(listener);

    service.sendMachineEvent(MachineEvent.iToValue(value: TestValue.bar()));

    await container.pump();
    expect(container.read(provider.service), service);
    verifyOnly(listener, listener(argThat(isAValueOf(TestValue.bar()))));
    expect(service.mounted, true);

    container.dispose();

    expect(service.mounted, false);
  });
}
